package org.xbib.interlibrary.statistics;

import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.xssf.usermodel.XSSFCellStyle;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.xbib.content.xml.util.XMLUtil;
import org.xbib.datastructures.csv.Generator;
import org.xbib.interlibrary.api.ApplicationModule;
import org.xbib.interlibrary.api.library.Role;
import org.xbib.interlibrary.api.order.statistics.Statistics;
import org.xbib.interlibrary.common.Interlibrary;
import org.xbib.interlibrary.common.InterlibraryModule;
import org.xbib.interlibrary.db.DatabaseModule;

public class StatisticsModule implements InterlibraryModule {

    private Interlibrary interlibrary;

    private DatabaseModule databaseModule;

    public StatisticsModule() {
    }

    @Override
    public String getName() {
        return "StatisticsModule";
    }

    @Override
    public List<Class<? extends ApplicationModule>> dependsOn() {
        return List.of(DatabaseModule.class);
    }

    @Override
    public void setInterlibrary(Interlibrary interlibrary) {
        this.interlibrary = interlibrary;
        this.databaseModule = interlibrary.getModule(DatabaseModule.class);
    }

    @Override
    public Interlibrary getInterlibrary() {
        return interlibrary;
    }

    @Override
    public void run() {
    }

    @Override
    public void close() throws IOException {
    }

    public enum Target {
        GENERATE,
        READ,
        CSV,
        XML,
        XLSX
    }

    private static final Logger logger = Logger.getLogger(StatisticsModule.class.getName());

    public List<String> getRoyaltiesPeriods(Role role,
                                            String isil) {
        return getRoyaltiesPeriods(role, isil, null, null);
    }

    public List<String> getRoyaltiesPeriods(Role role,
                                            String isil,
                                            LocalDateTime startDay,
                                            LocalDateTime endDay) {
        Map.Entry<LocalDateTime, LocalDateTime> minMax = null;
        if (isil != null) {
            if (startDay != null && endDay != null) {
                minMax = databaseModule.getStatisticsManager().findRoyaltiesStatisticsMinMax(role, startDay, endDay, isil);
            } else {
                minMax = databaseModule.getStatisticsManager().findRoyaltiesStatisticsMinMax(role, isil);
            }
        } else {
            if (startDay != null && endDay != null) {
                minMax = databaseModule.getStatisticsManager().findRoyaltiesStatisticsMinMax(role, startDay, endDay);
            } else {
                minMax = databaseModule.getStatisticsManager().findRoyaltiesStatisticsMinMax(role);
            }
        }
        List<String> list = new ArrayList<>();
        LocalDateTime min = minMax.getKey();
        // adjust min to lower limit
        LocalDate minDate = LocalDate.of(min.getYear(), min.getMonthValue() < 7 ? 1 : 7, 1);
        LocalDateTime max = minMax.getValue();
        // adjust max to upper limit
        LocalDate maxDate = max.getMonthValue() < 7 ?
                LocalDate.of(max.getYear(), 7, 1) : LocalDate.of(max.getYear() + 1, 1, 1);
        logger.log(Level.INFO, "min = " + min + " max = " + max + " mindate = " + minDate + " maxdate = " + maxDate);

        // min might equals to max, in that case, we have only a single statistics element
        while (minDate.isBefore(maxDate)) {
            list.add(minDate.getYear() + "-" + (minDate.getMonthValue() < 7 ? "1" : "2"));
            minDate = minDate.plusMonths(6L);
        }
        logger.log(Level.INFO, "list = " + list);
        return list;
    }

    public RoyaltiesStatistics getRoyalties(Role role,
                                            String isil,
                                            String consortium,
                                            String period,
                                            LocalDateTime start,
                                            LocalDateTime end,
                                            Target target,
                                            Path path) throws IOException {
        Statistics statistics = null;
        switch (target) {
            case GENERATE -> databaseModule.getStatisticsManager().putRoyaltiesStatistics(role, start, end);
            case READ -> statistics = isil != null ?
                    databaseModule.getStatisticsManager().getRoyaltiesStatistics(role, start, end, isil) :
                    databaseModule.getStatisticsManager().getRoyaltiesStatistics(role, start, end);
            case CSV -> {
                statistics = isil != null ?
                        databaseModule.getStatisticsManager().getRoyaltiesStatistics(role, start, end, isil) :
                        databaseModule.getStatisticsManager().getRoyaltiesStatistics(role, start, end);
                generateRoyaltiesCsv(statistics, path, consortium, period);
            }
            case XML -> {
                statistics = isil != null ?
                        databaseModule.getStatisticsManager().getRoyaltiesStatistics(role, start, end, isil) :
                        databaseModule.getStatisticsManager().getRoyaltiesStatistics(role, start, end);
                generateRoyaltiesXml(statistics, path, consortium, period);
            }
            case XLSX -> {
                statistics = isil != null ?
                        databaseModule.getStatisticsManager().getRoyaltiesStatistics(role, start, end, isil) :
                        databaseModule.getStatisticsManager().getRoyaltiesStatistics(role, start, end);
                generateRoyaltiesXslx(statistics, path, consortium,  period);
            }
            default -> throw new UnsupportedOperationException();
        }
        return new RoyaltiesStatistics(statistics);
    }

    public void changeRoyalties(Role role, String number, String value) {
        databaseModule.getStatisticsManager().putRoyaltiesStatisticsErrata(role, number, value);
    }

    private void generateRoyaltiesXml(Statistics statistics,
                                      Path path,
                                      String consortium,
                                      String period) throws IOException {
        Files.createDirectories(path.getParent());
        try (Writer writer = new OutputStreamWriter(Files.newOutputStream(path))) {
            writer.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
            writer.write("<records>\n");
            List<? extends Statistics.Entry> list = statistics.getEntries();
            for (int i = 0; i < list.size(); i++) {
                Statistics.Entry entry = list.get(i);
                writer.write("<record>\n");
                writer.write("<consortium>" + consortium + "</consortium>\n");
                writer.write("<period>" + period + "</period>\n");
                for (String column : statistics.getColumnNames()) {
                    List<String> elementNames = switch (column) {
                        case "ISIL" -> List.of("isil");
                        case "Nummer" -> List.of("consortid", "localid");
                        case "Titel" -> List.of("title");
                        case "Aufsatztitel" -> List.of("article");
                        case "Autor" -> List.of("author");
                        case "Verlag" -> List.of("publisher");
                        case "ISBN" -> List.of("isbn");
                        case "ISSN" -> List.of("issn");
                        case "Jahr" -> List.of("year");
                        case "Seiten" -> List.of("pages");
                        case "Modus" -> List.of("delivery");
                        default -> null;
                    };
                    if (elementNames != null) {
                        for (String elementName : elementNames) {
                            writer.write("<" + elementName + ">");
                            Object object = entry.get(column);
                            if (object != null) {
                                writer.write(XMLUtil.escape(XMLUtil.sanitizeXml10(object.toString())).toString());
                            }
                            writer.write("</" + elementName + ">\n");
                        }
                    }
                }
                writer.write("</record>\n");
            }
            writer.write("</records>\n");
        }
    }

    private void generateRoyaltiesCsv(Statistics statistics,
                                      Path path,
                                      String consortium,
                                      String period) throws IOException {
        Files.createDirectories(path.getParent());
        try (Writer writer = new OutputStreamWriter(Files.newOutputStream(path))) {
            Generator gen = new Generator(writer).setSeparator(',').setAlwaysQuote(true);
            List<String> keys = new ArrayList<>();
            keys.add("consortium");
            keys.add("period");
            for (String column : statistics.getColumnNames()) {
                List<String> elementNames = translateRoyaltiesColumn(column);
                if (elementNames != null) {
                    keys.addAll(elementNames);
                }
            }
            gen.keys(keys);
            gen.writeKeys();
            List<? extends Statistics.Entry> list = statistics.getEntries();
            for (int i = 0; i < list.size(); i++) {
                Statistics.Entry entry = list.get(i);
                gen.write(consortium);
                gen.write(period);
                for (String column : statistics.getColumnNames()) {
                    List<String> elementNames = translateRoyaltiesColumn(column);
                    if (elementNames != null) {
                        for (String elementName : elementNames) {
                            Object object = entry.get(column);
                            if (object != null) {
                                // we use XML escape here for compatibility with the XML format
                                gen.write(XMLUtil.escape(XMLUtil.sanitizeXml10(object.toString())).toString());
                            } else {
                                gen.write("");
                            }
                        }
                    }
                }
            }
            gen.close();
        }
    }

    private void generateRoyaltiesXslx(Statistics statistics,
                                       Path path,
                                       String consortium,
                                       String period) throws IOException {
        Files.createDirectories(path.getParent());
        try (OutputStream outputStream = Files.newOutputStream(path); XSSFWorkbook wb = new XSSFWorkbook()) {
            Sheet sheet = wb.createSheet(path.getFileName().toString());
            XSSFCellStyle locked = wb.createCellStyle();
            locked.setLocked(true);
            XSSFCellStyle unlocked = wb.createCellStyle();
            unlocked.setLocked(false);
            final Row header = sheet.createRow(0);
            List<String> keys = new ArrayList<>();
            keys.add("consortium");
            keys.add("period");
                for (String column : statistics.getColumnNames()) {
                    List<String> elementNames = translateRoyaltiesColumn(column);
                    if (elementNames != null) {
                        keys.addAll(elementNames);
                    }
                }
                int j = 0;
                for (String k : keys) {
                    Cell cell = header.createCell(j++);
                    cell.setCellValue(k);
                    cell.setCellStyle(locked);
                }
                List<? extends Statistics.Entry> list = statistics.getEntries();
                for (int i = 0; i < list.size(); i++) {
                    Statistics.Entry entry = list.get(i);
                    Row row = sheet.createRow(i + 1);
                    j = 0;
                    Cell cell;
                    cell = row.createCell(j++);
                    cell.setCellValue(consortium);
                    cell.setCellStyle(unlocked);
                    cell = row.createCell(j++);
                    cell.setCellValue(period);
                    cell.setCellStyle(unlocked);
                    for (String column : statistics.getColumnNames()) {
                        List<String> elementNames = translateRoyaltiesColumn(column);
                        if (elementNames != null) {
                            for (String elementName : elementNames) {
                                Object object = entry.get(column);
                                cell = row.createCell(j++);
                                if (object != null) {
                                    // we use XML escape here for compatibility with the XML format
                                    cell.setCellValue(XMLUtil.escape(XMLUtil.sanitizeXml10(object.toString())).toString());
                                    cell.setCellStyle(unlocked);
                                } else {
                                    cell.setCellValue("");
                                    cell.setCellStyle(unlocked);
                                }
                            }
                        }
                    }
                }
                sheet.createFreezePane(0, 1);
                wb.write(outputStream);
        }
    }

    private List<String> translateRoyaltiesColumn(String column) {
        return switch (column) {
            case "ISIL" -> List.of("isil");
            case "Nummer" -> List.of("consortid", "localid");
            case "Titel" -> List.of("title");
            case "Aufsatztitel" -> List.of("article");
            case "Autor" -> List.of("author");
            case "Verlag" -> List.of("publisher");
            case "ISBN" -> List.of("isbn");
            case "ISSN" -> List.of("issn");
            case "Jahr" -> List.of("year");
            case "Seiten" -> List.of("pages");
            case "Modus" -> List.of("delivery");
            default -> null;
        };
    }

    public void generateTitleStatistics(LocalDateTime start,
                                        LocalDateTime end) throws IOException {
        String from = start.format(DateTimeFormatter.ISO_LOCAL_DATE);
        String to = end.format(DateTimeFormatter.ISO_LOCAL_DATE);
        Statistics statistics = databaseModule.getStatisticsManager().getTitleStatistics(start, end);
        String prefix = "/var/tmp/stat/title/";
        Files.createDirectories(Paths.get(prefix));
        List<? extends Statistics.Entry> list = statistics.getEntries();
        for (int i = 0; i < list.size(); i++) {
            Statistics.Entry entry = list.get(i);
            String gb = (String) entry.get("Bibliothek");
            String activeFilename = prefix + gb + "-title-" + from + "-" + to + ".csv";
            try (Writer writer = new FileWriter(activeFilename, true)) {
                Generator gen = new Generator(writer).setSeparator(',').setAlwaysQuote(true);
                gen.keys(statistics.getColumnNames());
                for (String name : statistics.getColumnNames()) {
                    Object object = entry.get(name);
                    if (object != null) {
                        gen.write(object.toString());
                    } else {
                        gen.write("");
                    }
                }
            }
        }
    }
}
